@@ -25,51 +25,6 @@
 #include "automount.h"
 #include "nsswitch.h"
 
-static int check_nss_result(struct nss_source *this, enum nsswitch_status result)
-{
-	enum nsswitch_status status;
-	struct nss_action a;
-
-	/* Check if we have negated actions */
-	for (status = 0; status < NSS_STATUS_MAX; status++) {
-		a = this->action[status];
-		if (a.action == NSS_ACTION_UNKNOWN)
-			continue;
-
-		if (a.negated && result != status) {
-			if (a.action == NSS_ACTION_RETURN) {
-				if (result == NSS_STATUS_SUCCESS)
-					return 1;
-				else
-					return 0;
-			}
-		}
-	}
-
-	a = this->action[result];
-
-	/* Check if we have other actions for this status */
-	switch (result) {
-	case NSS_STATUS_SUCCESS:
-		if (a.action == NSS_ACTION_CONTINUE)
-			break;
-		return 1;
-
-	case NSS_STATUS_NOTFOUND:
-	case NSS_STATUS_UNAVAIL:
-	case NSS_STATUS_TRYAGAIN:
-		if (a.action == NSS_ACTION_RETURN) {
-			return 0;
-		}
-		break;
-
-	default:
-		break;
-	}
-
-	return -1;
-}
-
 static void nsslist_cleanup(void *arg)
 {
 	struct list_head *nsslist = (struct list_head *) arg;
@@ -89,9 +44,9 @@ static int do_read_master(struct master *master, char *type, time_t age)
 	argv[0] = master->name;
 	argv[1] = NULL;
 
-	lookup = open_lookup(type, "", NULL, argc, argv);
-	if (!lookup)
-		return NSS_STATUS_UNAVAIL;
+	status = open_lookup(type, "", NULL, argc, argv, &lookup);
+	if (status != NSS_STATUS_SUCCESS)
+		return status;
 
 	status = lookup->lookup_read_master(master, age, lookup->context);
 
@@ -220,7 +175,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 		if (result == NSS_STATUS_UNAVAIL)
 			master->read_fail = 1;
 
-		return !result;
+		return result;
 	} else {
 		char *name = master->name;
 		char *tmp;
@@ -270,7 +225,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 				if (result == NSS_STATUS_UNAVAIL)
 					master->read_fail = 1;
 
-				return !result;
+				return result;
 			}
 		}
 	}
@@ -282,10 +237,11 @@ int lookup_nss_read_master(struct master *master, time_t age)
 		if (!list_empty(&nsslist))
 			free_sources(&nsslist);
 		error(logopt, "can't to read name service switch config.");
-		return 0;
+		return NSS_STATUS_UNAVAIL;
 	}
 
 	/* First one gets it */
+	result = NSS_STATUS_SUCCESS;
 	head = &nsslist;
 	list_for_each(p, head) {
 		struct nss_source *this;
@@ -293,6 +249,13 @@ int lookup_nss_read_master(struct master *master, time_t age)
 
 		this = list_entry(p, struct nss_source, list);
 
+		if (strncmp(this->source, "files", 5) &&
+		    strncmp(this->source, "nis", 3) &&
+		    strncmp(this->source, "nisplus", 7) &&
+		    strncmp(this->source, "ldap", 4) &&
+		    strncmp(this->source, "sss", 3))
+			continue;
+
 		debug(logopt,
 		      "reading master %s %s", this->source, master->name);
 
@@ -319,8 +282,16 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			}
 		}
 
-		if (result == NSS_STATUS_UNKNOWN) {
+		/* We've been instructed to move onto the next source */
+		if (result == NSS_STATUS_TRYAGAIN) {
+			result = NSS_STATUS_SUCCESS;
+			continue;
+		}
+
+		if (result == NSS_STATUS_UNKNOWN ||
+		    result == NSS_STATUS_NOTFOUND) {
 			debug(logopt, "no map - continuing to next source");
+			result = NSS_STATUS_SUCCESS;
 			continue;
 		}
 
@@ -337,7 +308,7 @@ int lookup_nss_read_master(struct master *master, time_t age)
 	if (!list_empty(&nsslist))
 		free_sources(&nsslist);
 
-	return !result;
+	return result;
 }
 
 static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t age)
@@ -345,16 +316,27 @@ static int do_read_map(struct autofs_point *ap, struct map_source *map, time_t a
 	struct lookup_mod *lookup;
 	int status;
 
-	lookup = open_lookup(map->type, "", map->format, map->argc, map->argv);
-	if (!lookup) {
-		debug(ap->logopt, "lookup module %s failed", map->type);
-		return NSS_STATUS_UNAVAIL;
-	}
-
+	lookup = NULL;
 	master_source_writelock(ap->entry);
-	if (map->lookup)
-		close_lookup(map->lookup);
-	map->lookup = lookup;
+	if (!map->lookup) {
+		status = open_lookup(map->type, "", map->format,
+				     map->argc, map->argv, &lookup);
+		if (status != NSS_STATUS_SUCCESS) {
+			master_source_unlock(ap->entry);
+			debug(ap->logopt,
+			      "lookup module %s open failed", map->type);
+			return status;
+		}
+		map->lookup = lookup;
+	} else {
+		lookup = map->lookup;
+		status = lookup->lookup_reinit(map->format,
+					       map->argc, map->argv,
+					       &lookup->context);
+		if (status)
+			warn(ap->logopt,
+			     "lookup module %s reinit failed", map->type);
+	}
 	master_source_unlock(ap->entry);
 
 	if (!map->stale)
@@ -681,6 +663,56 @@ int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time
 	return 0;
 }
 
+static char *make_browse_path(unsigned int logopt,
+			      const char *root, const char *key,
+			      const char *prefix)
+{
+	unsigned int l_prefix;
+	unsigned int k_len, r_len;
+	char *k_start;
+	char *path;
+
+	k_start = (char *) key;
+	k_len = strlen(key);
+	l_prefix = 0;
+
+	if (prefix) {
+		l_prefix = strlen(prefix);
+
+		if (l_prefix > k_len)
+			return NULL;
+
+		/* If the prefix doesn't match the beginning
+		 * of the key this entry isn't a sub directory
+		 * at this level.
+		 */
+		if (strncmp(key, prefix, l_prefix))
+			return NULL;
+
+		/* Directory entry starts following the prefix */
+		k_start += l_prefix;
+	}
+
+	/* No remaining "/" allowed here */
+	if (strchr(k_start, '/'))
+		return NULL;
+
+	r_len = strlen(root);
+
+	if ((r_len + strlen(k_start)) > KEY_MAX_LEN)
+		return NULL;
+
+	path = malloc(r_len + k_len + 2);
+	if (!path) {
+		warn(logopt, "failed to allocate full path");
+		return NULL;
+	}
+
+	sprintf(path, "%s/%s", root, k_start);
+
+	return path;
+}
+
 int lookup_ghost(struct autofs_point *ap, const char *root)
 {
 	struct master_mapent *entry = ap->entry;
@@ -724,10 +756,19 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 			if (!me->mapent)
 				goto next;
 
-			if (!strcmp(me->key, "*"))
+			/* Wildcard cannot be a browse directory and amd map
+			 * keys may end with the wildcard.
+			 */
+			if (strchr(me->key, '*'))
 				goto next;
 
+			/* This will also take care of amd "/defaults" entry as
+			 * amd map keys are not allowd to start with "/"
+			 */
 			if (*me->key == '/') {
+				if (map->flags & MAP_FLAG_FORMAT_AMD)
+					goto next;
+
 				/* It's a busy multi-mount - leave till next time */
 				if (list_empty(&me->multi_list))
 					error(ap->logopt,
@@ -735,12 +776,10 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 				goto next;
 			}
 
-			fullpath = malloc(strlen(me->key) + strlen(root) + 3);
-			if (!fullpath) {
-				warn(ap->logopt, "failed to allocate full path");
+			fullpath = make_browse_path(ap->logopt,
+						    root, me->key, ap->pref);
+			if (!fullpath)
 				goto next;
-			}
-			sprintf(fullpath, "%s/%s", root, me->key);
 
 			ret = stat(fullpath, &st);
 			if (ret == -1 && errno != ENOENT) {
@@ -750,6 +789,17 @@ int lookup_ghost(struct autofs_point *ap, const char *root)
 				goto next;
 			}
 
+			/* Directory already exists? */
+			if (!ret) {
+				/* Shouldn't need this
+				me->dev = st.st_dev;
+				me->ino = st.st_ino;
+				*/
+				debug(ap->logopt, "me->dev %d me->ino %d", me->dev, me->ino);
+				free(fullpath);
+				goto next;
+			}
+
 			ret = mkdir_path(fullpath, 0555);
 			if (ret < 0 && errno != EEXIST) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
@@ -782,12 +832,12 @@ int do_lookup_mount(struct autofs_point *ap, struct map_source *map, const char
 	int status;
 
 	if (!map->lookup) {
-		lookup = open_lookup(map->type, "",
-				     map->format, map->argc, map->argv);
-		if (!lookup) {
+		status = open_lookup(map->type, "",
+				     map->format, map->argc, map->argv, &lookup);
+		if (status != NSS_STATUS_SUCCESS) {
 			debug(ap->logopt,
-			      "lookup module %s failed", map->type);
-			return NSS_STATUS_UNAVAIL;
+			      "lookup module %s open failed", map->type);
+			return status;
 		}
 		map->lookup = lookup;
 	}
@@ -820,7 +870,7 @@ static int lookup_amd_instance(struct autofs_point *ap,
 		return NSS_STATUS_UNKNOWN;
 	}
 
-	m_key = malloc(strlen(ap->path) + strlen(me->multi->key) + 1);
+	m_key = malloc(strlen(ap->path) + strlen(me->multi->key) + 2);
 	if (!m_key) {
 		error(ap->logopt,
 		     "failed to allocate storage for search key");
@@ -875,7 +925,7 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 	struct map_source *instance;
 	char src_file[] = "file";
 	char src_prog[] = "program";
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	struct stat st;
 	char *type, *format;
 
@@ -915,7 +965,7 @@ static int lookup_name_source_instance(struct autofs_point *ap, struct map_sourc
 {
 	struct map_source *instance;
 	const char *format;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 
 	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
 		return lookup_amd_instance(ap, map, name, name_len);
@@ -1070,8 +1120,15 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 		 */
 		cache_unlock(me->mc);
 	else {
-		/* Notify only once after fail */
-		logmsg("key \"%s\" not found in map source(s).", name);
+		if (!defaults_disable_not_found_message()) {
+			/* This really should be a warning but the original
+			 * request for this needed it to be unconditional.
+			 * That produces, IMHO, unnecessary noise in the log
+			 * so a configuration option has been added to provide
+			 * the ability to turn it off.
+			 */
+			logmsg("key \"%s\" not found in map source(s).", name);
+		}
 
 		/* Doesn't exist in any source, just add it somewhere */
 		if (source)
@@ -1079,7 +1136,7 @@ static void update_negative_cache(struct autofs_point *ap, struct map_source *so
 		else
 			map = entry->maps;
 		if (map) {
-			time_t now = time(NULL);
+			time_t now = monotonic_time(NULL);
 			int rv = CHE_FAIL;
 
 			cache_writelock(map->mc);
@@ -1107,7 +1164,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 	struct nss_source *this;
 	struct map_source *map;
 	enum nsswitch_status status;
-	int result = 0;
+	int result = NSS_STATUS_UNKNOWN;
 
 	/*
 	 * For each map source (ie. each entry for the mount
@@ -1126,6 +1183,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 		 * the map entry was last updated.
 		 */
 		if (entry->age > map->age) {
+			status = NSS_STATUS_UNAVAIL;
 			map = map->next;
 			continue;
 		}
@@ -1147,6 +1205,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 				char *tmp = strdup("ldap");
 				if (!tmp) {
 					map = map->next;
+					status = NSS_STATUS_TRYAGAIN;
 					continue;
 				}
 				map->type = tmp;
@@ -1179,6 +1238,7 @@ int lookup_nss_mount(struct autofs_point *ap, struct map_source *source, const c
 			    !strcmp(this->source, "sss")) {
 				warn(ap->logopt,
 				     "source sss is not available for amd maps.");
+				result = NSS_STATUS_UNAVAIL;
 				continue;
 			}
 
@@ -1249,28 +1309,23 @@ void lookup_close_lookup(struct autofs_point *ap)
 	return;
 }
 
-static char *make_fullpath(const char *root, const char *key)
+static char *make_fullpath(struct autofs_point *ap, const char *key)
 {
+	char *path = NULL;
 	int l;
-	char *path;
 
-	if (*key == '/') {
+	if (*key != '/')
+		path = make_browse_path(ap->logopt, ap->path, key, ap->pref);
+	else {
 		l = strlen(key) + 1;
 		if (l > KEY_MAX_LEN)
-			return NULL;
+			goto out;
 		path = malloc(l);
 		if (!path)
-			return NULL;
+			goto out;
 		strcpy(path, key);
-	} else {
-		l = strlen(key) + 1 + strlen(root) + 1;
-		if (l > KEY_MAX_LEN)
-			return NULL;
-		path = malloc(l);
-		if (!path)
-			return NULL;
-		sprintf(path, "%s/%s", root, key);
 	}
+out:
 	return path;
 }
 
@@ -1301,13 +1356,14 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 
 		key = strdup(me->key);
 		me = cache_enumerate(mc, me);
-		if (!key || !strcmp(key, "*")) {
+		/* Don't consider any entries with a wildcard */
+		if (!key || strchr(key, '*')) {
 			if (key)
 				free(key);
 			continue;
 		}
 
-		path = make_fullpath(ap->path, key);
+		path = make_fullpath(ap, key);
 		if (!path) {
 			warn(ap->logopt, "can't malloc storage for path");
 			free(key);
