@@ -30,6 +30,9 @@
 
 #define MAPFMT_DEFAULT "sun"
 
+/* Half a second between retries */
+#define SETAUTOMOUNTENT_MASTER_INTERVAL	500000000
+
 #define MODPREFIX "lookup(sss): "
 
 #define SSS_SO_NAME "libsss_autofs"
@@ -56,38 +59,16 @@ struct lookup_context {
 
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
 
-int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
+static int open_sss_lib(struct lookup_context *ctxt)
 {
-	struct lookup_context *ctxt;
-	char buf[MAX_ERR_BUF];
 	char dlbuf[PATH_MAX];
 	char *estr;
 	void *dh;
 	size_t size;
 
-	*context = NULL;
-
-	ctxt = malloc(sizeof(struct lookup_context));
-	if (!ctxt) {
-		estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-
-	if (argc < 1) {
-		free(ctxt);
-		logerr(MODPREFIX "No map name");
-		return 1;
-	}
-	ctxt->mapname = argv[0];
-
-	if (!mapfmt)
-		mapfmt = MAPFMT_DEFAULT;
-
 	size = snprintf(dlbuf, sizeof(dlbuf),
 			"%s/%s.so", SSS_LIB_DIR, SSS_SO_NAME);
 	if (size >= sizeof(dlbuf)) {
-		free(ctxt);
 		logmsg(MODPREFIX "sss library path too long");
 		return 1;
 	}
@@ -95,7 +76,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	dh = dlopen(dlbuf, RTLD_LAZY);
 	if (!dh) {
 		logerr(MODPREFIX "failed to open %s: %s", dlbuf, dlerror());
-		free(ctxt);
 		return 1;
 	}
 	ctxt->dlhandle = dh;
@@ -116,15 +96,6 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	if (!ctxt->setautomntent)
 		goto lib_names_fail;
 
-	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
-	if (!ctxt->parse) {
-		logmsg(MODPREFIX "failed to open parse context");
-		dlclose(dh);
-		free(ctxt);
-		return 1;
-	}
-	*context = ctxt;
-
 	return 0;
 
 lib_names_fail:
@@ -133,10 +104,109 @@ lib_names_fail:
 	else
 		logerr(MODPREFIX "dlsym: %s", estr);
 	dlclose(dh);
-	free(ctxt);
+
 	return 1;
 }
 
+static int do_init(const char *mapfmt,
+		   int argc, const char *const *argv,
+		   struct lookup_context *ctxt, unsigned int reinit)
+{
+	int ret = 0;
+
+	if (argc < 1) {
+		logerr(MODPREFIX "No map name");
+		ret = 1;
+		goto out;
+	}
+	ctxt->mapname = argv[0];
+
+	if (!mapfmt)
+		mapfmt = MAPFMT_DEFAULT;
+
+	if (!reinit) {
+		ret = open_sss_lib(ctxt);
+		if (ret)
+			goto out;
+	}
+
+	if (reinit) {
+		ret = reinit_parse(ctxt->parse, mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (ret)
+			logmsg(MODPREFIX "failed to reinit parse context");
+	} else {
+		ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
+		if (!ctxt->parse) {
+			logmsg(MODPREFIX "failed to open parse context");
+			dlclose(ctxt->dlhandle);
+			ret = 1;
+		}
+	}
+out:
+	return ret;
+}
+
+int lookup_init(const char *mapfmt,
+		int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt;
+	char buf[MAX_ERR_BUF];
+	char *estr;
+
+	*context = NULL;
+
+	ctxt = malloc(sizeof(struct lookup_context));
+	if (!ctxt) {
+		estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	if (do_init(mapfmt, argc, argv, ctxt, 0)) {
+		free(ctxt);
+		return 1;
+	}
+
+	*context = ctxt;
+
+	return 0;
+}
+
+int lookup_reinit(const char *mapfmt,
+		  int argc, const char *const *argv, void **context)
+{
+	struct lookup_context *ctxt = (struct lookup_context *) *context;
+	struct lookup_context *new;
+	char buf[MAX_ERR_BUF];
+	int ret;
+
+	new = malloc(sizeof(struct lookup_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	new->parse = ctxt->parse;
+	ret = do_init(mapfmt, argc, argv, new, 1);
+	if (ret) {
+		free(new);
+		return 1;
+	}
+
+	new->dlhandle = ctxt->dlhandle;
+	new->setautomntent = ctxt->setautomntent;
+	new->getautomntent_r = ctxt->getautomntent_r;
+	new->getautomntbyname_r = ctxt->getautomntbyname_r;
+	new->endautomntent = ctxt->endautomntent;
+
+	*context = new;
+
+	free(ctxt);
+
+	return 0;
+}
+
 static int setautomntent(unsigned int logopt,
 			 struct lookup_context *ctxt, const char *mapname,
 			 void **sss_ctxt)
@@ -148,9 +218,55 @@ static int setautomntent(unsigned int logopt,
 		error(logopt, MODPREFIX "setautomntent: %s", estr);
 		if (*sss_ctxt)
 			free(*sss_ctxt);
-		return 0;
 	}
-	return 1;
+	return ret;
+}
+
+static int setautomntent_wait(unsigned int logopt,
+			      struct lookup_context *ctxt,
+			      const char *mapname,
+			      void **sss_ctxt, unsigned int retries)
+{
+	unsigned int retry = 0;
+	int ret = 0;
+
+	*sss_ctxt = NULL;
+
+	while (++retry < retries) {
+		struct timespec t = { 0, SETAUTOMOUNTENT_MASTER_INTERVAL };
+		struct timespec r;
+
+		ret = ctxt->setautomntent(mapname, sss_ctxt);
+		if (ret != ENOENT)
+			break;
+
+		if (*sss_ctxt) {
+			free(*sss_ctxt);
+			*sss_ctxt = NULL;
+		}
+
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+	}
+
+
+	if (ret) {
+		char buf[MAX_ERR_BUF];
+		char *estr;
+
+		if (*sss_ctxt) {
+			free(*sss_ctxt);
+			*sss_ctxt = NULL;
+		}
+
+		if (retry == retries)
+			ret = ETIMEDOUT;
+
+		estr = strerror_r(ret, buf, MAX_ERR_BUF);
+		error(logopt, MODPREFIX "setautomntent: %s", estr);
+	}
+
+	return ret;
 }
 
 static int endautomntent(unsigned int logopt,
@@ -161,9 +277,8 @@ static int endautomntent(unsigned int logopt,
 		char buf[MAX_ERR_BUF];
 		char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
 		error(logopt, MODPREFIX "endautomntent: %s", estr);
-		return 0;
 	}
-	return 1;
+	return ret;
 }
 
 int lookup_read_master(struct master *master, time_t age, void *context)
@@ -180,8 +295,26 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	char *value = NULL;
 	int count, ret;
 
-	if (!setautomntent(logopt, ctxt, ctxt->mapname, &sss_ctxt))
-		return NSS_STATUS_UNAVAIL;
+	ret = setautomntent(logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		unsigned int retries;
+
+		if (ret != ENOENT)
+			return NSS_STATUS_UNAVAIL;
+
+		retries = defaults_get_sss_master_map_wait() * 2;
+		if (retries <= 0)
+			return NSS_STATUS_NOTFOUND;
+
+		ret = setautomntent_wait(logopt,
+					 ctxt, ctxt->mapname, &sss_ctxt,
+					 retries);
+		if (ret) {
+			if (ret == ENOENT)
+				return NSS_STATUS_NOTFOUND;
+			return NSS_STATUS_UNAVAIL;
+		}
+	}
 
 	count = 0;
 	while (1) {
@@ -271,17 +404,23 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	mc = source->mc;
 
 	/*
-	 * If we don't need to create directories then there's no use
-	 * reading the map. We always need to read the whole map for
-	 * direct mounts in order to mount the triggers.
+	 * If we don't need to create directories (or don't need
+	 * to read an amd cache:=all map) then there's no use
+	 * reading the map. We always need to read the whole map
+	 * for direct mounts in order to mount the triggers.
 	 */
-	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT) {
+	if (ap->type != LKP_DIRECT &&
+	    !(ap->flags & (MOUNT_FLAG_GHOST|MOUNT_FLAG_AMD_CACHE_ALL))) {
 		debug(ap->logopt, "map read not needed, so not done");
 		return NSS_STATUS_SUCCESS;
 	}
 
-	if (!setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt))
+	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		if (ret == ENOENT)
+			return NSS_STATUS_NOTFOUND;
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	count = 0;
 	while (1) {
@@ -374,7 +513,7 @@ static int lookup_one(struct autofs_point *ap,
 	struct mapent_cache *mc;
 	struct mapent *we;
 	void *sss_ctxt = NULL;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	char buf[MAX_ERR_BUF];
 	char *value = NULL;
 	char *s_key;
@@ -386,8 +525,12 @@ static int lookup_one(struct autofs_point *ap,
 
 	mc = source->mc;
 
-	if (!setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt))
+	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	if (ret) {
+		if (ret == ENOENT)
+			return NSS_STATUS_NOTFOUND;
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	ret = ctxt->getautomntbyname_r(qKey, &value, sss_ctxt);
 	if (ret && ret != ENOENT) {
@@ -491,7 +634,7 @@ static int check_map_indirect(struct autofs_point *ap,
 	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *me;
-	time_t now = time(NULL);
+	time_t now = monotonic_time(NULL);
 	time_t t_last_read;
 	int ret, cur_state;
 
@@ -585,7 +728,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	/* Check if we recorded a mount fail for this key anywhere */
 	me = lookup_source_mapent(ap, key, LKP_DISTINCT);
 	if (me) {
-		if (me->status >= time(NULL)) {
+		if (me->status >= monotonic_time(NULL)) {
 			cache_unlock(me->mc);
 			return NSS_STATUS_NOTFOUND;
 		} else {
