@@ -232,27 +232,15 @@ int expandsunent(const char *src, char *dst, const char *key,
 	return len;
 }
 
-int parse_init(int argc, const char *const *argv, void **context)
+static int do_init(int argc, const char *const *argv, struct parse_context *ctxt)
 {
-	struct parse_context *ctxt;
-	char buf[MAX_ERR_BUF];
 	char *noptstr, *def, *val, *macros, *gbl_options;
-	const char *xopt;
+	char buf[MAX_ERR_BUF];
 	int optlen, len, offset;
+	const char *xopt;
 	int i, bval;
 	unsigned int append_options;
 
-	/* Set up context and escape chain */
-
-	if (!(ctxt = (struct parse_context *) malloc(sizeof(struct parse_context)))) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		*context = NULL;
-		return 1;
-	}
-	*context = (void *) ctxt;
-
-	*ctxt = default_context;
 	optlen = 0;
 
 	/* Look for options and capture, and create new defines if we need to */
@@ -357,9 +345,7 @@ int parse_init(int argc, const char *const *argv, void **context)
 			}
 			if (!noptstr) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				kill_context(ctxt);
 				logerr(MODPREFIX "%s", estr);
-				*context = NULL;
 				return 1;
 			}
 			ctxt->optstr = noptstr;
@@ -391,9 +377,36 @@ int parse_init(int argc, const char *const *argv, void **context)
 		}
 	}
 options_done:
+
 	debug(LOGOPT_NONE,
 	      MODPREFIX "init gathered global options: %s", ctxt->optstr);
 
+	return 0;
+}
+
+int parse_init(int argc, const char *const *argv, void **context)
+{
+	struct parse_context *ctxt;
+	char buf[MAX_ERR_BUF];
+
+	*context = NULL;
+
+	/* Set up context and escape chain */
+
+	ctxt = (struct parse_context *) malloc(sizeof(struct parse_context));
+	if (!ctxt) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	*ctxt = default_context;
+
+	if (do_init(argc, argv, ctxt)) {
+		free(ctxt);
+		return 1;
+	}
+
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
 	instance_mutex_lock();
@@ -404,12 +417,39 @@ options_done:
 			init_ctr++;
 		} else {
 			kill_context(ctxt);
-			*context = NULL;
 			instance_mutex_unlock();
 			return 1;
 		}
 	}
 	instance_mutex_unlock();
+
+	*context = (void *) ctxt;
+
+	return 0;
+}
+
+int parse_reinit(int argc, const char *const *argv, void **context)
+{
+	struct parse_context *ctxt = (struct parse_context *) *context;
+	struct parse_context *new;
+	char buf[MAX_ERR_BUF];
+
+	new = (struct parse_context *) malloc(sizeof(struct parse_context));
+	if (!new) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		return 1;
+	}
+
+	*new = default_context;
+
+	if (do_init(argc, argv, new))
+		return 1;
+
+	kill_context(ctxt);
+
+	*context = (void *) new;
+
 	return 0;
 }
 
@@ -1194,6 +1234,29 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 	return rv;
 }
 
+static char *do_expandsunent(const char *src, const char *key,
+			     const struct substvar *svc, int slashify_colons)
+{
+	char *mapent;
+	int len;
+
+	len = expandsunent(src, NULL, key, svc, slashify_colons);
+	if (len == 0) {
+		errno = EINVAL;
+		return NULL;
+	}
+	len++;
+
+	mapent = malloc(len);
+	if (!mapent)
+		return NULL;
+	memset(mapent, 0, len);
+
+	expandsunent(src, mapent, key, svc, slashify_colons);
+
+	return mapent;
+}
+
 /*
  * syntax is:
  *	[-options] location [location] ...
@@ -1233,34 +1296,48 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	}
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	macro_lock();
 
+	/* Offset map entries have been expanded already, avoid expanding
+	 * them again so that the quote handling is consistent between map
+	 * entry locations and (previously expanded) offset map entry
+	 * locations.
+	 */
+	if (*name == '/') {
+		cache_readlock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (me && me->multi && me->multi != me) {
+			cache_unlock(mc);
+			mapent_len = strlen(mapent) + 1;
+			pmapent = malloc(mapent_len + 1);
+			if (!pmapent) {
+				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				logerr(MODPREFIX "malloc: %s", estr);
+				return 1;
+			}
+			memset(pmapent, 0, mapent_len + 1);
+			strcpy(pmapent, mapent);
+			goto dont_expand;
+		}
+		cache_unlock(mc);
+	}
+
+	macro_lock();
 	ctxt->subst = addstdenv(ctxt->subst, NULL);
 
-	mapent_len = expandsunent(mapent, NULL, name, ctxt->subst, slashify);
-	if (mapent_len == 0) {
+	pmapent = do_expandsunent(mapent, name, ctxt->subst, slashify);
+	if (!pmapent) {
 		error(ap->logopt, MODPREFIX "failed to expand map entry");
 		ctxt->subst = removestdenv(ctxt->subst, NULL);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
 	}
+	mapent_len = strlen(pmapent) + 1;
 
-	pmapent = alloca(mapent_len + 1);
-	if (!pmapent) {	
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "alloca: %s", estr);
-		ctxt->subst = removestdenv(ctxt->subst, NULL);
-		macro_unlock();
-		pthread_setcancelstate(cur_state, NULL);
-		return 1;
-	}
-	pmapent[mapent_len] = '\0';
-
-	expandsunent(mapent, pmapent, name, ctxt->subst, slashify);
 	ctxt->subst = removestdenv(ctxt->subst, NULL);
-
 	macro_unlock();
+
+dont_expand:
 	pthread_setcancelstate(cur_state, NULL);
 
 	debug(ap->logopt, MODPREFIX "expanded entry: %s", pmapent);
@@ -1269,6 +1346,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	options = strdup(ctxt->optstr ? ctxt->optstr : "");
 	if (!options) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		free(pmapent);
 		logerr(MODPREFIX "strdup: %s", estr);
 		return 1;
 	}
@@ -1297,6 +1375,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 					if (mnt_options)
 						free(mnt_options);
 					free(options);
+					free(pmapent);
 					return 1;
 				}
 				mnt_options = tmp;
@@ -1323,6 +1402,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 						free(options);
 					if (mnt_options)
 						free(mnt_options);
+					free(pmapent);
 					return 1;
 				}
 				options = tmp;
@@ -1346,6 +1426,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
+				free(pmapent);
 				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
@@ -1356,6 +1437,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
+				free(pmapent);
 				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
@@ -1369,6 +1451,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		me = cache_lookup_distinct(mc, name);
 		if (!me) {
 			free(options);
+			free(pmapent);
 			cache_unlock(mc);
 			pthread_setcancelstate(cur_state, NULL);
 			error(ap->logopt,
@@ -1400,6 +1483,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			 */
 			if (source->flags & MAP_FLAG_FORMAT_AMD) {
 				free(options);
+				free(pmapent);
 				cache_multi_unlock(me);
 				cache_unlock(mc);
 				pthread_setcancelstate(cur_state, NULL);
@@ -1432,6 +1516,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				cache_multi_unlock(me);
 				cache_unlock(mc);
 				free(options);
+				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
@@ -1446,6 +1531,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				cache_unlock(mc);
 				free(path);
 				free(options);
+				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
@@ -1467,6 +1553,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				cache_unlock(mc);
 				free(path);
 				free(options);
+				free(pmapent);
 				free(myoptions);
 				if (loc)
 					free(loc);
@@ -1495,6 +1582,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		cache_unlock(mc);
 
 		free(options);
+		free(pmapent);
 		pthread_setcancelstate(cur_state, NULL);
 
 		return rv;
@@ -1515,6 +1603,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			loc = strdup(p);
 			if (!loc) {
 				free(options);
+				free(pmapent);
 				cache_unlock(mc);
 				warn(ap->logopt, MODPREFIX "out of memory");
 				return 1;
@@ -1525,6 +1614,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			cache_unlock(mc);
 			free(loc);
 			free(options);
+			free(pmapent);
 			return rv;
 		}
 		cache_unlock(mc);
@@ -1533,6 +1623,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		loc = dequote(p, l, ap->logopt);
 		if (!loc) {
 			free(options);
+			free(pmapent);
 			warn(ap->logopt, MODPREFIX "null location or out of memory");
 			return 1;
 		}
@@ -1540,6 +1631,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		/* Location can't begin with a '/' */
 		if (*p == '/') {
 			free(options);
+			free(pmapent);
 			free(loc);
 			warn(ap->logopt,
 			      MODPREFIX "error location begins with \"/\"");
@@ -1549,6 +1641,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		if (!validate_location(ap->logopt, loc)) {
 			free(loc);
 			free(options);
+			free(pmapent);
 			return 1;
 		}
 
@@ -1566,6 +1659,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (!ent) {
 				free(loc);
 				free(options);
+				free(pmapent);
 				warn(ap->logopt,
 				     MODPREFIX "null location or out of memory");
 				return 1;
@@ -1575,6 +1669,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				free(ent);
 				free(loc);
 				free(options);
+				free(pmapent);
 				return 1;
 			}
 
@@ -1586,6 +1681,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				free(ent);
 				free(loc);
 				free(options);
+				free(pmapent);
 				error(ap->logopt, MODPREFIX "out of memory");
 				return 1;
 			}
@@ -1616,6 +1712,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			if (loclen == 0) {
 				free(loc);
 				free(options);
+				free(pmapent);
 				error(ap->logopt,
 				      MODPREFIX "entry %s is empty!", name);
 				return 1;
@@ -1636,6 +1733,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		if (loc)
 			free(loc);
 		free(options);
+		free(pmapent);
 		pthread_setcancelstate(cur_state, NULL);
 	}
 	return rv;
